--- printf_fixes_3.c	2007-10-03 16:25:05.000000000 -0600
+++ ide_atapi_changes_4.c	2007-10-03 16:00:15.000000000 -0600
@@ -147,6 +147,7 @@
 #define LINUXBIOS	1
 
 #define BX_USE_ATADRV    1
+#define EDD_USE_VERSION_3 0
 #define BX_ELTORITO_BOOT 1
 
 #define BX_MAX_ATA_INTERFACES   2
@@ -718,6 +719,8 @@
   //     The EBDA structure should conform to
   //     http://www.frontiernet.net/~fys/rombios.htm document
   //     I made the ata and cdemu structs begin at 0x121 in the EBDA seg
+  // EBDA must be at most 768 bytes; it lives at 0x9fc00, and the boot
+  // device tables are at 0x9ff00 -- 0x9ffff
   typedef struct {
     unsigned char filler1[0x3D];
 
@@ -762,8 +765,8 @@
     Bit32u  sector_count1;
     Bit32u  sector_count2;
     Bit16u  blksize;
-    Bit16u  dpte_segment;
     Bit16u  dpte_offset;
+    Bit16u  dpte_segment;
     Bit16u  key;
     Bit8u   dpi_length;
     Bit8u   reserved1;
@@ -904,11 +907,15 @@
 static void           print_cdromboot_failure();
 
 # if BX_USE_ATADRV
+#define udelay() (inb(0x80)||inb(0x80))
 
 // ATA / ATAPI driver
+void   mdelay();
 void   ata_init();
 void   ata_detect();
 void   ata_reset();
+Bit16u   ata_device_selection();
+void   ata_device_reset();
 
 Bit16u ata_cmd_non_data();
 Bit16u ata_cmd_data_in();
@@ -2160,6 +2167,73 @@
   write_byte(ebda_seg,&EbdaData->ata.cdcount,0);
 }
 
+#define TIMEOUT 0
+#define BSY 1
+#define NOT_BSY 2
+#define NOT_BSY_DRQ 3
+#define NOT_BSY_NOT_DRQ 4
+#define NOT_BSY_RDY 5
+
+#if LINUXBIOS // !!FIXME!! How do you tell if you're on real hardware?
+  #define IDE_TIMEOUT 32000u
+#else
+  #define IDE_TIMEOUT 32u
+#endif
+
+void print_status();
+static void print_status(base)
+           Bit16u base;
+{
+  BX_DEBUG_ATA("\tstatus=%2x, err=%2x, sc = %2x, cl = %2x ch = %2x dh=%2x\n",
+                inb(base+ATA_CB_STAT), inb(base+ATA_CB_ERR), inb(base+ATA_CB_SC),
+                inb(base+ATA_CB_CL), inb(base+ATA_CB_CH), inb(base+ATA_CB_DH));
+}
+
+int await_ide();
+static int await_ide(when_done,base,timeout)
+  Bit8u when_done;
+  Bit16u base;
+  Bit16u timeout;
+{
+  Bit32u time=0,last=0;
+  Bit16u status;
+  Bit8u result;
+  status = inb(base + ATA_CB_STAT); // for the times you're supposed to throw one away
+  for(;;) {
+    status = inb(base+ATA_CB_STAT);
+    time++;
+    if (when_done == BSY)
+      result = status & ATA_CB_STAT_BSY;
+    else if (when_done == NOT_BSY)
+      result = !(status & ATA_CB_STAT_BSY);
+    else if (when_done == NOT_BSY_DRQ)
+      result = !(status & ATA_CB_STAT_BSY) && (status & ATA_CB_STAT_DRQ);
+    else if (when_done == NOT_BSY_NOT_DRQ)
+      result = !(status & ATA_CB_STAT_BSY) && !(status & ATA_CB_STAT_DRQ);
+    else if (when_done == NOT_BSY_RDY)
+      result = !(status & ATA_CB_STAT_BSY) && (status & ATA_CB_STAT_RDY);
+    else if (when_done == TIMEOUT)
+      result = 0;
+
+    if (result) return 0;
+    if (time>>16 != last) // mod 2048 each 16 ms
+    {
+      last = time >>16;
+      BX_DEBUG_ATA("await_ide: (TIMEOUT,BSY,!BSY,!BSY_DRQ,!BSY_!DRQ) %d time= %ld timeout= %d\n",when_done,time>>11, timeout);
+      print_status(base);
+    }
+    if (status & ATA_CB_STAT_ERR)
+    {
+      BX_DEBUG_ATA("await_ide: ERROR (TIMEOUT,BSY,!BSY,!BSY_DRQ,!BSY_!DRQ) %d time= %ld timeout= %d\n",when_done,time>>11, timeout);
+      print_status(base);
+      return -1;
+    }
+    if ((timeout == 0) || ((time>>11) > timeout)) break;
+  }
+  BX_INFO("IDE time out\n");
+  return -1;
+}
+
 // ---------------------------------------------------------------------------
 // ATA/ATAPI driver : device detection
 // ---------------------------------------------------------------------------
@@ -2169,7 +2243,6 @@
   Bit16u ebda_seg=read_word(0x0040,0x000E);
   Bit8u  hdcount, cdcount, device, type;
   Bit8u  buffer[0x0200];
-  Bit16u i;
 
 #if BX_MAX_ATA_INTERFACES > 0
   write_byte(ebda_seg,&EbdaData->ata.channels[0].iface,ATA_IFACE_ISA);
@@ -2204,7 +2277,7 @@
 
   for(device=0; device<BX_MAX_ATA_DEVICES; device++) {
     Bit16u iobase1, iobase2;
-    Bit8u  channel, slave, shift;
+    Bit8u  channel, slave;
     Bit8u  sc, sn, cl, ch, st;
 
     channel = device / 2;
@@ -2233,7 +2306,22 @@
       write_byte(ebda_seg,&EbdaData->ata.devices[device].type,ATA_TYPE_UNKNOWN);
 
       // reset the channel
-      ata_reset(device);
+        // ATA-4
+        // 9.3.1-2 Software reset - Device 0 or 1
+        // 9.3.1 (a), 9.3.2 (a) -- set SRST in DC
+        outb(iobase2+ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN | ATA_CB_DC_SRST);
+        udelay(); udelay(); udelay(); udelay(); udelay(); //5 us
+
+        // 9.3.1 (c), 9.3.2 (c) -- wait for BSY
+        await_ide(BSY, iobase1, IDE_TIMEOUT);
+
+        // 9.3.1 (f), 9.3.2 (g) -- clear SRST
+        outb(iobase2+ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN);
+
+        // 9.3.1 (l), 9.3.2 (m) -- wait for !BSY
+        await_ide(NOT_BSY, iobase1, IDE_TIMEOUT);
+
+        // 9.3.1 (m), 9.3.2 (n) -- wait for !BSY and DRDY if not ATA_TYPE_ATAPI
 
       // check for ATA or ATAPI
       outb(iobase1+ATA_CB_DH, slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0);
@@ -2247,12 +2335,23 @@
         if ((cl==0x14) && (ch==0xeb)) {
           write_byte(ebda_seg,&EbdaData->ata.devices[device].type,ATA_TYPE_ATAPI);
         } else if ((cl==0x00) && (ch==0x00) && (st!=0x00)) {
+          await_ide(NOT_BSY_RDY, iobase1, IDE_TIMEOUT);
           write_byte(ebda_seg,&EbdaData->ata.devices[device].type,ATA_TYPE_ATA);
         } else if ((cl==0xff) && (ch==0xff)) {
+          await_ide(NOT_BSY_RDY, iobase1, IDE_TIMEOUT);
           write_byte(ebda_seg,&EbdaData->ata.devices[device].type,ATA_TYPE_NONE);
         }
       }
     }
+    // Enable Interrupts
+    outb(iobase2+ATA_CB_DC, ATA_CB_DC_HD15);
+   }
+
+  for(device=0; device<BX_MAX_ATA_DEVICES; device++) {
+    Bit8u  channel, slave, shift;
+
+    channel = device / 2;
+    slave = device % 2;
 
     type=read_byte(ebda_seg,&EbdaData->ata.devices[device].type);
 
@@ -2453,7 +2552,7 @@
 // ATA/ATAPI driver : software reset
 // ---------------------------------------------------------------------------
 // ATA-3
-// 8.2.1 Software reset - Device 0
+// 8.2.1 Software reset - Device 0 or 1
 
 void   ata_reset(device)
 Bit16u device;
@@ -2461,6 +2560,7 @@
   Bit16u ebda_seg=read_word(0x0040,0x000E);
   Bit16u iobase1, iobase2;
   Bit8u  channel, slave, sn, sc;
+  Bit8u  ch, cl;
   Bit16u max;
 
   channel = device / 2;
@@ -2471,47 +2571,68 @@
 
   // Reset
 
-// 8.2.1 (a) -- set SRST in DC
+// 9.3.1 (a), 9.3.2 (a) -- set SRST in DC
   outb(iobase2+ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN | ATA_CB_DC_SRST);
+  udelay(); udelay(); udelay(); udelay(); udelay(); //5 us
 
-// 8.2.1 (b) -- wait for BSY
-  max=0xff;
-  while(--max>0) {
-    Bit8u status = inb(iobase1+ATA_CB_STAT);
-    if ((status & ATA_CB_STAT_BSY) != 0) break;
-  }
+// 9.3.1 (c), 9.3.2 (c) -- wait for BSY
+  await_ide(BSY, iobase1, IDE_TIMEOUT);
 
-// 8.2.1 (f) -- clear SRST
+// 9.3.1 (f), 9.3.2 (g) -- clear SRST
   outb(iobase2+ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN);
 
-  if (read_byte(ebda_seg,&EbdaData->ata.devices[device].type) != ATA_TYPE_NONE) {
-
-// 8.2.1 (g) -- check for sc==sn==0x01
+// 9.3.1 (j), 9.3.2 (k) -- check for sc==sn
     // select device
-    outb(iobase1+ATA_CB_DH, slave?ATA_CB_DH_DEV1:ATA_CB_DH_DEV0);
-    sc = inb(iobase1+ATA_CB_SC);
-    sn = inb(iobase1+ATA_CB_SN);
+  outb(iobase1+ATA_CB_DH, slave?ATA_CB_DH_DEV1:ATA_CB_DH_DEV0);
+  sc = inb(iobase1+ATA_CB_SC);
+  sn = inb(iobase1+ATA_CB_SN);
+
+  if ( (sc==0x01) && (sn==0x01) ) {
+    ch = inb(iobase1+ATA_CB_CH);
+    cl = inb(iobase1+ATA_CB_CL);
+// 9.3.1 (l), 9.3.2 (m) -- wait for !BSY
+// 9.3.1 (m), 9.3.2 (n) -- wait for !BSY and RDY if not ATA_TYPE_ATAPI
+    if (cl == 0 && ch == 0) //ATA
+      await_ide(NOT_BSY_RDY, iobase1, IDE_TIMEOUT);
+    else //ATAPI
+      await_ide(NOT_BSY, iobase1, IDE_TIMEOUT);
+  }
+}
 
-    if ( (sc==0x01) && (sn==0x01) ) {
+//ATA-4 9.6 Device Selection protocol
+// Figure 11
+// BSY=0 and DRQ=0 before starting
 
-// 8.2.1 (h) -- wait for not BSY
-      max=0xff;
-      while(--max>0) {
-        Bit8u status = inb(iobase1+ATA_CB_STAT);
-        if ((status & ATA_CB_STAT_BSY) == 0) break;
-        }
-      }
-    }
+Bit16u   ata_device_selection(device,base)
+Bit16u device,base;
+{
+  Bit16u status;
+  return await_ide(NOT_BSY_NOT_DRQ, base, IDE_TIMEOUT);
+  outb(base + ATA_CB_DH, (device%2 ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0));
+  udelay(); /* spec says wait 400 ns , 1 us is close enough */
+  return await_ide(NOT_BSY_NOT_DRQ, base, IDE_TIMEOUT);
+}
 
-// 8.2.1 (i) -- wait for DRDY
-  max=0xfff;
-  while(--max>0) {
-    Bit8u status = inb(iobase1+ATA_CB_STAT);
-      if ((status & ATA_CB_STAT_RDY) != 0) break;
-  }
+//ATA-4 9.4 DEVICE RESET protocol
+// The whole thing should take < 6 s
 
-  // Enable interrupts
-  outb(iobase2+ATA_CB_DC, ATA_CB_DC_HD15);
+void   ata_device_reset(device)
+Bit16u device;
+{
+  Bit16u status;
+  Bit16u base;
+  Bit16u ebda_seg=read_word(0x0040,0x000E);
+  Bit8u slave = device & 0x1;
+  base = read_word(ebda_seg, &EbdaData->ata.channels[device/2].iobase1);
+
+  //9.4 (a)
+  outb(base + ATA_CB_DH, (slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0));
+  outb(base + ATA_CB_CMD, ATA_CMD_DEVICE_RESET);
+  //9.4 (c)
+  udelay(); /* 9.4 (c) spec says wait 400 ns , 1 us is close enough */
+  await_ide(BSY, base, IDE_TIMEOUT); // This could be shorter
+  //9.4 (j)
+  await_ide(NOT_BSY, base, IDE_TIMEOUT);
 }
 
 // ---------------------------------------------------------------------------
@@ -2552,22 +2673,20 @@
   if (mode == ATA_MODE_PIO32) blksize>>=2;
   else blksize>>=1;
 
+  //!!FIXME!! this only works up to 28-bit LBA
   // sector will be 0 only on lba access. Convert to lba-chs
   if (sector == 0) {
     sector = (Bit16u) (lba & 0x000000ffL);
-    lba >>= 8;
-    cylinder = (Bit16u) (lba & 0x0000ffffL);
-    lba >>= 16;
-    head = ((Bit16u) (lba & 0x0000000fL)) | 0x40;
-    }
+    cylinder = (Bit16u) ((lba>>8) & 0x0000ffffL);
+    head = ((Bit16u) ((lba>>24) & 0x0000000fL)) | ATA_CB_DH_LBA;
+  }
 
   // Reset count of transferred data
   write_word(ebda_seg, &EbdaData->ata.trsfsectors,0);
   write_dword(ebda_seg, &EbdaData->ata.trsfbytes,0L);
   current = 0;
 
-  status = inb(iobase1 + ATA_CB_STAT);
-  if (status & ATA_CB_STAT_BSY) return 1;
+  if (ata_device_selection(device,iobase1)!=0) return 1;
 
   outb(iobase2 + ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN);
   outb(iobase1 + ATA_CB_FR, 0x00);
@@ -2578,13 +2697,16 @@
   outb(iobase1 + ATA_CB_DH, (slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0) | (Bit8u) head );
   outb(iobase1 + ATA_CB_CMD, command);
 
-  while (1) {
-    status = inb(iobase1 + ATA_CB_STAT);
-    if ( !(status & ATA_CB_STAT_BSY) ) break;
-    }
+  udelay(); /* spec says wait 400 ns , 1 us is close enough */
+  await_ide(NOT_BSY_DRQ, iobase1, IDE_TIMEOUT);
+  status = inb(iobase1 + ATA_CB_STAT);
 
+  //BX_INFO("ata_cmd_data_in: device %d command %x count %x cylinder %x head %x sector %x lba %lx current %x\n",device, command, count, cylinder, head, sector, lba, current);
+  //BX_INFO("ata_cmd_data_in: seg %x off %x first_dword %lx\n",segment, offset, read_dword(segment,offset));
   if (status & ATA_CB_STAT_ERR) {
-    BX_DEBUG_ATA("ata_cmd_data_in : read error\n");
+    BX_DEBUG_ATA("ata_cmd_data_in : read error count = %x, sector = %x, cylinder = %x, head = %x, command = %x\n",
+                  count, sector, cylinder, head, command);
+    print_status(iobase1);
     return 2;
     } else if ( !(status & ATA_CB_STAT_DRQ) ) {
     BX_DEBUG_ATA("ata_cmd_data_in : DRQ not set (status %02x)\n", (unsigned) status);
@@ -2641,6 +2763,7 @@
     current++;
     write_word(ebda_seg, &EbdaData->ata.trsfsectors,current);
     count--;
+    await_ide(NOT_BSY,iobase1,IDE_TIMEOUT);
     status = inb(iobase1 + ATA_CB_STAT);
     if (count == 0) {
       if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ | ATA_CB_STAT_ERR) )
@@ -2653,7 +2776,8 @@
     else {
       if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ | ATA_CB_STAT_ERR) )
           != (ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ) ) {
-        BX_DEBUG_ATA("ata_cmd_data_in : more sectors left (status %02x)\n", (unsigned) status);
+        BX_DEBUG_ATA("ata_cmd_data_in : more sectors left (status %02x)\n", status);
+        BX_DEBUG_ATA("ata_cmd_data_in : device %d command %x count %x cylinder %x head %x sector %x lba %lx segment %x offset %x current %x\n",device, command, count, cylinder, head, sector, lba, segment, offset, current);
         return 5;
       }
       continue;
@@ -2698,19 +2822,16 @@
   // sector will be 0 only on lba access. Convert to lba-chs
   if (sector == 0) {
     sector = (Bit16u) (lba & 0x000000ffL);
-    lba >>= 8;
-    cylinder = (Bit16u) (lba & 0x0000ffffL);
-    lba >>= 16;
-    head = ((Bit16u) (lba & 0x0000000fL)) | 0x40;
-    }
+    cylinder = (Bit16u) ((lba>>8) & 0x0000ffffL);
+    head = ((Bit16u) ((lba>>24) & 0x0000000fL)) | ATA_CB_DH_LBA;
+  }
 
   // Reset count of transferred data
   write_word(ebda_seg, &EbdaData->ata.trsfsectors,0);
   write_dword(ebda_seg, &EbdaData->ata.trsfbytes,0L);
   current = 0;
 
-  status = inb(iobase1 + ATA_CB_STAT);
-  if (status & ATA_CB_STAT_BSY) return 1;
+  if (ata_device_selection(device,iobase1)!=0) return 1;
 
   outb(iobase2 + ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN);
   outb(iobase1 + ATA_CB_FR, 0x00);
@@ -2721,10 +2842,9 @@
   outb(iobase1 + ATA_CB_DH, (slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0) | (Bit8u) head );
   outb(iobase1 + ATA_CB_CMD, command);
 
-  while (1) {
-    status = inb(iobase1 + ATA_CB_STAT);
-    if ( !(status & ATA_CB_STAT_BSY) ) break;
-    }
+  udelay(); /* spec says wait 400 ns , 1 us is close enough */
+  await_ide(NOT_BSY_DRQ, iobase1, IDE_TIMEOUT);
+  status = inb(iobase1 + ATA_CB_STAT);
 
   if (status & ATA_CB_STAT_ERR) {
     BX_DEBUG_ATA("ata_cmd_data_out : read error\n");
@@ -2859,23 +2979,20 @@
   write_word(ebda_seg, &EbdaData->ata.trsfsectors,0);
   write_dword(ebda_seg, &EbdaData->ata.trsfbytes,0L);
 
-  status = inb(iobase1 + ATA_CB_STAT);
-  if (status & ATA_CB_STAT_BSY) return 2;
+  if (ata_device_selection(device,iobase1)!=0) return 2;
 
   outb(iobase2 + ATA_CB_DC, ATA_CB_DC_HD15 | ATA_CB_DC_NIEN);
-  // outb(iobase1 + ATA_CB_FR, 0x00);
-  // outb(iobase1 + ATA_CB_SC, 0x00);
-  // outb(iobase1 + ATA_CB_SN, 0x00);
+  outb(iobase1 + ATA_CB_FR, 0x00);
+  outb(iobase1 + ATA_CB_SC, 0x00);
+  outb(iobase1 + ATA_CB_SN, 0x00);
   outb(iobase1 + ATA_CB_CL, 0xfff0 & 0x00ff);
   outb(iobase1 + ATA_CB_CH, 0xfff0 >> 8);
   outb(iobase1 + ATA_CB_DH, slave ? ATA_CB_DH_DEV1 : ATA_CB_DH_DEV0);
   outb(iobase1 + ATA_CB_CMD, ATA_CMD_PACKET);
 
-  // Device should ok to receive command
-  while (1) {
-    status = inb(iobase1 + ATA_CB_STAT);
-    if ( !(status & ATA_CB_STAT_BSY) ) break;
-    }
+  udelay(); /* spec says wait 400 ns , 1 us is close enough */
+  await_ide(NOT_BSY_DRQ, iobase1, IDE_TIMEOUT);
+  status = inb(iobase1 + ATA_CB_STAT);
 
   if (status & ATA_CB_STAT_ERR) {
     BX_DEBUG_ATA("ata_cmd_packet : error, status is %02x\n",status);
@@ -2911,28 +3028,35 @@
 ASM_END
 
   if (inout == ATA_DATA_NO) {
+    await_ide(NOT_BSY, iobase1, IDE_TIMEOUT);
     status = inb(iobase1 + ATA_CB_STAT);
     }
   else {
+        Bit16u loops = 0;
+        Bit8u  err, sc;
   while (1) {
 
+      status = inb(iobase2 + ATA_CB_ASTAT);
+      if (loops == 0) //first time through
+        await_ide(NOT_BSY_DRQ, iobase1, IDE_TIMEOUT);
+      else
+        await_ide(NOT_BSY, iobase1, IDE_TIMEOUT);
+      loops++;
+
       status = inb(iobase1 + ATA_CB_STAT);
+      sc = inb(iobase1 + ATA_CB_SC);
+
+      BX_DEBUG_ATA("ata_cmd_packet: loop %d status=%02x sc=%02x\n", loops,status,sc);
 
       // Check if command completed
-      if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_DRQ) ) ==0 ) break;
+      if(((inb(iobase1 + ATA_CB_SC)&0x7)==0x3) && 
+         ((status & (ATA_CB_STAT_RDY | ATA_CB_STAT_ERR)) == ATA_CB_STAT_RDY)) break;
 
       if (status & ATA_CB_STAT_ERR) {
         BX_DEBUG_ATA("ata_cmd_packet : error (status %02x)\n",status);
         return 3;
       }
 
-      // Device must be ready to send data
-      if ( (status & (ATA_CB_STAT_BSY | ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ | ATA_CB_STAT_ERR) )
-            != (ATA_CB_STAT_RDY | ATA_CB_STAT_DRQ) ) {
-        BX_DEBUG_ATA("ata_cmd_packet : not ready (status %02x)\n", status);
-        return 4;
-        }
-
       // Normalize address
       bufseg += (bufoff / 16);
       bufoff %= 16;
@@ -3091,51 +3215,107 @@
 // ---------------------------------------------------------------------------
 
   Bit16u
-atapi_get_sense(device)
+atapi_get_sense(device, seg, asc, ascq)
   Bit16u device;
 {
   Bit8u  atacmd[12];
-  Bit8u  buffer[16];
+  Bit8u  buffer[18];
   Bit8u i;
 
   memsetb(get_SS(),atacmd,0,12);
 
   // Request SENSE
-  atacmd[0]=0x03;
-  atacmd[4]=0x20;
-  if (ata_cmd_packet(device, 12, get_SS(), atacmd, 0, 16L, ATA_DATA_IN, get_SS(), buffer) != 0)
+  atacmd[0]=ATA_CMD_REQUEST_SENSE;
+  atacmd[4]=sizeof(buffer);
+  if (ata_cmd_packet(device, 12, get_SS(), atacmd, 0, 18L, ATA_DATA_IN, get_SS(), buffer) != 0)
     return 0x0002;
 
-  if ((buffer[0] & 0x7e) == 0x70) {
-    return (((Bit16u)buffer[2]&0x0f)*0x100)+buffer[12];
-    }
+  write_byte(seg,asc,buffer[12]);
+  write_byte(seg,ascq,buffer[13]);
 
   return 0;
 }
 
   Bit16u
-atapi_is_ready(device)
+atapi_is_ready(device) //detect medium from FILO
   Bit16u device;
 {
-  Bit8u  atacmd[12];
-  Bit8u  buffer[];
-
-  memsetb(get_SS(),atacmd,0,12);
+	Bit8u packet[12];
+	Bit8u buf[8];
+	Bit32u block_len;
+	Bit32u sectors;
+	Bit32u timeout; //measured in ms
+	Bit32u time;
+	Bit8u asc, ascq;
+	Bit8u in_progress;
+	Bit16u ebda_seg = read_word(0x0040,0x000E);
+	if (read_byte(ebda_seg,&EbdaData->ata.devices[device].type) != ATA_TYPE_ATAPI) {
+		printf("not implemented for non-ATAPI device\n");
+		return -1;
+	}
+
+	BX_DEBUG_ATA("ata_detect_medium: begin\n");
+	print_status(read_word(ebda_seg,&EbdaData->ata.channels[device/2].iobase1));
+	memsetb(get_SS(),packet, 0, sizeof packet);
+	packet[0] = 0x25; /* READ CAPACITY */
+
+	/* Retry READ CAPACITY for 5 seconds unless MEDIUM NOT PRESENT
+         * is reported by the device. If the device reports "IN PROGRESS",
+         * 30 seconds is added. */
+	timeout = 5000;
+	time = 0;
+	in_progress = 0;
+	while (time < timeout) {
+		if (ata_cmd_packet(device, sizeof(packet), get_SS(), packet, 0, 8L, ATA_DATA_IN, get_SS(), buf) == 0)
+			goto ok;
+
+		if (atapi_get_sense(device, get_SS(), &asc, &ascq) == 0) {
+			if (asc == 0x3a) { /* MEDIUM NOT PRESENT */
+				BX_DEBUG_ATA("Device reports MEDIUM NOT PRESENT\n");
+				return -1;
+			}
+
+			if (asc == 0x04 && ascq == 0x01 && !in_progress) {
+					/* IN PROGRESS OF BECOMING READY */
+				printf("Waiting for device to detect medium... ");
+				/* Allow 30 seconds more */
+				timeout = 30000;
+				in_progress = 1;
+			}
+		}
+
+		mdelay(1000);
+		time += 100;
+	}
+	BX_DEBUG_ATA("read capacity failed\n");
+	return -1;
+ok:
+
+	block_len = (Bit32u) buf[4] << 24
+		| (Bit32u) buf[5] << 16
+		| (Bit32u) buf[6] << 8
+		| (Bit32u) buf[7] << 0;
+	BX_DEBUG_ATA("block_len=%u\n", block_len);
 
-  // Test Unit Ready
-  if (ata_cmd_packet(device, 12, get_SS(), atacmd, 0, 0L, ATA_DATA_NO, get_SS(), buffer) != 0)
-    return 0x000f;
-
-  if (atapi_get_sense(device) !=0 ) {
-    memsetb(get_SS(),atacmd,0,12);
-
-    // try to send Test Unit Ready again
-    if (ata_cmd_packet(device, 12, get_SS(), atacmd, 0, 0L, ATA_DATA_NO, get_SS(), buffer) != 0)
-      return 0x000f;
+           if (block_len!= 2048 && block_len!= 512)
+           {
+		printf("Unsupported sector size %u\n", block_len);
+		return -1;
+           }
+	write_dword(ebda_seg,&EbdaData->ata.devices[device].blksize, block_len);
 
-    return atapi_get_sense(device);
-    }
-  return 0;
+	sectors = (Bit32u) buf[0] << 24
+		| (Bit32u) buf[1] << 16
+		| (Bit32u) buf[2] << 8
+		| (Bit32u) buf[3] << 0;
+
+	BX_DEBUG_ATA("sectors=%u\n", sectors);
+	if (block_len == 2048)
+		sectors <<= 2; /* # of sectors in 512-byte "soft" sector */
+	if (sectors != read_dword(ebda_seg,&EbdaData->ata.devices[device].sectors))
+		printf("%dMB medium detected\n", sectors>>(20-9));
+	write_dword(ebda_seg,&EbdaData->ata.devices[device].sectors, sectors);
+	return 0;
 }
 
   Bit16u
@@ -3168,6 +3348,17 @@
 // Start of El-Torito boot functions
 // ---------------------------------------------------------------------------
 
+// I read somewhere that one inb = 1 us, but it seems more like .5 us
+  void
+mdelay(msecs)
+Bit16u msecs;
+{
+  Bit16u i,j;
+  for (i=0;i<msecs;i++)
+    for (j=0;j<2048;j++)
+      udelay();
+}
+
   void
 cdemu_init()
 {
@@ -3212,6 +3403,9 @@
     if (atapi_is_cdrom(device)) break;
     }
 
+  if(error = atapi_is_ready(device) != 0)
+	bios_printf(BIOS_PRINTF_INFO | BIOS_PRINTF_SCREEN, "ata_is_ready returned %d\n",error);
+
   // if not found
   if(device >= BX_MAX_ATA_DEVICES) return 2;
 
@@ -4793,6 +4987,11 @@
       goto int13_success;
       break;
 
+    case 0x0d: /* alternate disk reset */
+      ata_device_reset(device);
+      goto int13_success;
+      break;
+
     case 0x01: /* read disk status */
       status = read_byte(0x0040, 0x0074);
       SET_AH(status);
@@ -4833,11 +5032,7 @@
       // FIXME verify
       if ( GET_AH() == 0x04 ) goto int13_success;
 
-      nph   = read_word(ebda_seg, &EbdaData->ata.devices[device].pchs.heads);
-      npspt = read_word(ebda_seg, &EbdaData->ata.devices[device].pchs.spt);
-
-      // if needed, translate lchs to lba, and execute command
-      if ( (nph != nlh) || (npspt != nlspt)) {
+      if (read_byte(ebda_seg, &EbdaData->ata.devices[device].translation)==ATA_TRANSLATION_LBA) {
         lba = ((((Bit32u)cylinder * (Bit32u)nlh) + (Bit32u)head) * (Bit32u)nlspt) + (Bit32u)sector - 1;
         sector = 0; // this forces the command to be lba
         }
@@ -4901,13 +5096,13 @@
 
     case 0x15: /* read disk drive size */
 
-      // Get physical geometry from table
-      npc   = read_word(ebda_seg, &EbdaData->ata.devices[device].pchs.cylinders);
-      nph   = read_word(ebda_seg, &EbdaData->ata.devices[device].pchs.heads);
-      npspt = read_word(ebda_seg, &EbdaData->ata.devices[device].pchs.spt);
+      // Get logical geometry from table
+      nlc   = read_word(ebda_seg, &EbdaData->ata.devices[device].lchs.cylinders);
+      nlh   = read_word(ebda_seg, &EbdaData->ata.devices[device].lchs.heads);
+      nlspt = read_word(ebda_seg, &EbdaData->ata.devices[device].lchs.spt);
 
       // Compute sector count seen by int13
-      lba = (Bit32u)(npc - 1) * (Bit32u)nph * (Bit32u)npspt;
+      lba = (Bit32u)(nlc - 1) * (Bit32u)nlh * (Bit32u)nlspt;
       CX = lba >> 16;
       DX = lba & 0xffff;
 
@@ -4995,8 +5190,16 @@
         blksize = read_word(ebda_seg, &EbdaData->ata.devices[device].blksize);
 
         write_word(DS, SI+(Bit16u)&Int13DPT->size, 0x1a);
-        write_word(DS, SI+(Bit16u)&Int13DPT->infos, 0x02); // geometry is valid
-        write_dword(DS, SI+(Bit16u)&Int13DPT->cylinders, (Bit32u)npc);
+        if ((lba/npspt)/nph > 0x3fff)
+        {
+          write_word(DS, SI+(Bit16u)&Int13DPT->infos, 0x00); // geometry is invalid
+          write_dword(DS, SI+(Bit16u)&Int13DPT->cylinders, 0x3fff);
+        }
+        else
+        {
+          write_word(DS, SI+(Bit16u)&Int13DPT->infos, 0x02); // geometry is valid
+          write_dword(DS, SI+(Bit16u)&Int13DPT->cylinders, (Bit32u)npc);
+        }
         write_dword(DS, SI+(Bit16u)&Int13DPT->heads, (Bit32u)nph);
         write_dword(DS, SI+(Bit16u)&Int13DPT->spt, (Bit32u)npspt);
         write_dword(DS, SI+(Bit16u)&Int13DPT->sector_count1, lba);  // FIXME should be Bit64
@@ -5022,14 +5225,14 @@
         mode = read_byte(ebda_seg, &EbdaData->ata.devices[device].mode);
         translation = read_byte(ebda_seg, &EbdaData->ata.devices[device].translation);
 
-        options  = (translation==ATA_TRANSLATION_NONE?0:1<<3); // chs translation
+        options  = (translation==ATA_TRANSLATION_NONE?0:1)<<3; // chs translation
         options |= (1<<4); // lba translation
-        options |= (mode==ATA_MODE_PIO32?1:0<<7);
-        options |= (translation==ATA_TRANSLATION_LBA?1:0<<9);
-        options |= (translation==ATA_TRANSLATION_RECHS?3:0<<9);
+        options |= (mode==ATA_MODE_PIO32?1:0)<<7;
+        options |= (translation==ATA_TRANSLATION_LBA?1:0)<<9;
+        options |= (translation==ATA_TRANSLATION_RECHS?3:0)<<9;
 
         write_word(ebda_seg, &EbdaData->ata.dpte.iobase1, iobase1);
-        write_word(ebda_seg, &EbdaData->ata.dpte.iobase2, iobase2);
+        write_word(ebda_seg, &EbdaData->ata.dpte.iobase2, iobase2 + ATA_CB_DC);
         write_byte(ebda_seg, &EbdaData->ata.dpte.prefix, (0xe | (device % 2))<<4 );
         write_byte(ebda_seg, &EbdaData->ata.dpte.unused, 0xcb );
         write_byte(ebda_seg, &EbdaData->ata.dpte.irq, irq );
@@ -5038,16 +5241,19 @@
         write_byte(ebda_seg, &EbdaData->ata.dpte.pio, 0 );
         write_word(ebda_seg, &EbdaData->ata.dpte.options, options);
         write_word(ebda_seg, &EbdaData->ata.dpte.reserved, 0);
-        write_byte(ebda_seg, &EbdaData->ata.dpte.revision, 0x11);
+        if (EDD_USE_VERSION_3 && size >=0x42)
+          write_byte(ebda_seg, &EbdaData->ata.dpte.revision, 0x11);
+        else
+          write_byte(ebda_seg, &EbdaData->ata.dpte.revision, 0x10);
 
         checksum=0;
-        for (i=0; i<15; i++) checksum+=read_byte(ebda_seg, (&EbdaData->ata.dpte) + i);
+        for (i=0; i<15; i++) checksum+=read_byte(ebda_seg, ((Bit8u*)(&EbdaData->ata.dpte)) + i);
         checksum = ~checksum;
         write_byte(ebda_seg, &EbdaData->ata.dpte.checksum, checksum);
         }
 
       // EDD 3.x
-      if(size >= 0x42) {
+      if(EDD_USE_VERSION_3 && size >= 0x42) {
         Bit8u channel, iface, checksum, i;
         Bit16u iobase1;
 
@@ -5113,7 +5319,6 @@
 
     case 0x09: /* initialize drive parameters */
     case 0x0c: /* seek to specified cylinder */
-    case 0x0d: /* alternate disk reset */
     case 0x11: /* recalibrate */
     case 0x14: /* controller internal diagnostic */
       BX_INFO("int13h_harddisk function %02xh unimplemented, returns success\n", GET_AH());
@@ -5377,7 +5582,7 @@
         options |= (mode==ATA_MODE_PIO32?1:0<<7);
 
         write_word(ebda_seg, &EbdaData->ata.dpte.iobase1, iobase1);
-        write_word(ebda_seg, &EbdaData->ata.dpte.iobase2, iobase2);
+        write_word(ebda_seg, &EbdaData->ata.dpte.iobase2, iobase2 + ATA_CB_DC);
         write_byte(ebda_seg, &EbdaData->ata.dpte.prefix, (0xe | (device % 2))<<4 );
         write_byte(ebda_seg, &EbdaData->ata.dpte.unused, 0xcb );
         write_byte(ebda_seg, &EbdaData->ata.dpte.irq, irq );
